// Arduino UNO用Scratch対応ファームウエア
// 注意：専用のヘルパーアプリが必要です
// by Takao Takahashi
// https://tsq.gapj.net
// このスケッチはI2Cを使用するためA4,A5は直接は使えません
// ScratchからアナログA4,A5を読み出しても値は取得できません

//PROGMEM使用
#include <avr/pgmspace.h> 
// MPL3115温度・気圧センサを使用する(I2C)
#include <Wire.h>
#include <Adafruit_MPL3115A2.h>
//サーボモータを使用する
#include <Servo.h>
// 液晶を使用するためのライブラリ
#include <Adafruit_RGBLCDShield.h>
#include <utility/Adafruit_MCP23017.h>
//気温・気圧センサのクラス定義
Adafruit_MPL3115A2 baro = Adafruit_MPL3115A2();
// 液晶のクラス定義
Adafruit_RGBLCDShield lcd = Adafruit_RGBLCDShield();
// サーボのクラス定義(2個まで)
Servo myServo1;
Servo myServo2;

//Arduino UNOのピン定義
static uint8_t digitalPins[] = {2,3,4,5,6,7,8,9,10,11,12,13}; //デジタルピン列挙
static uint16_t analogPins[] ={0,1,2,3}; //アナログピン列挙
//ピンの動作中モードを覚えておく
uint8_t skipmePins[16]; //analogWriteしたピンは読み出さないためのモードを覚えておく
                        //(PWM出力したピンをdigitalRead()すると出力が停止するため)

//コマンド文字列とその値用
//ヘルパーからの電文はA1 255のような形なので32バイトもあれば十分
char sbuffer[32];
//コマンドから指定されるピンとその値
uint8_t cmd_pin, cmd_value;
//ピン列挙の配列のサイズ格納用
uint8_t a_maxDigital;
uint8_t a_maxAnalog;

//気温と気圧
float my_temp;
float my_press;

//初期化処理
void setup() {
  //シリアルポート初期化
  Serial.begin(115200);
  Serial.flush();
  my_temp = (float) 0;
  my_press = (float) 0;
  //ピン配列のサイズをセット
  a_maxDigital = sizeof(digitalPins) / sizeof(uint8_t);
  a_maxAnalog = sizeof(analogPins) / sizeof(uint16_t);
}


//メインループ
void loop() {
  reportAllPins(); //全ピン読み出してシリアルへ送出
  sreadCommand();  //ヘルパーからのコマンド処理
}
//メインループはここまで

//ピンの値の読み込みとヘルパーアプリへの送信
//Scratchのリポーターブロックが使う値を送信する
void reportAllPins(){
  uint8_t i;
  uint16_t value;
  uint8_t a_max;

  //デジタルピンの読み込み
  for(i=0;i<a_maxDigital;i++){
    if(skipmePins[i] != 'A'){ //そのピンがアナログ(PWM)モードなら読み出さない
      value = digitalRead(digitalPins[i]);
    }
    Serial.print("D");
    Serial.print(digitalPins[i]);
    Serial.print(" ");
    Serial.print(value);
    Serial.print(":");
  }
  //アナログピンの読み込み
  //A/D変換の取り込み
  //Arduinoでは処理せずヘルパー(リポーターブロック)へ送る
  for(i=0;i<a_maxAnalog;i++){
    value = analogRead(analogPins[i]);
    Serial.print("A");
    Serial.print(analogPins[i]);
    Serial.print(" ");
    Serial.print(value);
    Serial.print(":");
  }
  //気温(T)と気圧(P)を送出
  Serial.print("T");
  Serial.print(" ");
  Serial.print(my_temp);
  Serial.print(":");
  Serial.print("P");
  Serial.print(" ");
  Serial.print(my_press);
  Serial.print("\n");
}

//シリアルからのコマンド受信
//ヘルパーからの動作指示を処理する
void sreadCommand(){
  uint8_t i;
  char tmp_c;

  if(Serial.available()){ //受信した文字があれば処理する
     i = 0;
     while(1){ //文字列受信完了までのループ
       if(Serial.available()){ //受信した文字がある
         tmp_c = Serial.read(); //1文字読み出し
         if(tmp_c == 0x0a) break; //行末はCRかLF
         if(tmp_c == 0x0d) break; //CRかLFを受信したら1行完了
         sbuffer[i] = tmp_c;
         sbuffer[i+1] = 0x00; //ひとつ先を0x00にしておいてバッファ初期化をさぼる
         i++;
         if(i>30) break; //終端(CRorLF)されず30文字越えたらやめる
       }
     }
     //受信したコマンドの処理
     //コマンドの文字(頭の1バイト)によって処理を切り替える
     switch(sbuffer[0]){
       case 'D': //デジタル出力
         digitalOut();
         break;
       case 'A': //アナログ出力
         analogOut();
         break;
       case 'M': //ピンモード設定
         pinmSet();
         break;
       case 'T': //気温・気圧取得
         getTempPress(); //処理に0.5秒以上かかかるので注意
         break;
       case 'U': //サーボ設定
         setServo();
         break;
       case 'V': //サーボ角度
         servoAngle();
         break;
       case 'W': //LCD初期化
         lcdInit();
         break;
       case 'X': //LCDバックライト制御
         lcdBacklight();
         break;
       case 'Y': //LCDにメッセージ表示
         lcdMessage();
         break;
//       case 'Z': //LCD表示クリア
//         lcdClear();
//         break;
       default : //該当するコマンドがなければ何もしない
         break;
     }
   } //何も受信していなければそのまま戻る
}


//デジタルポートへの出力処理
//High/Lowのデジタル出力
void digitalOut()
{
  uint8_t i;
  
  cmd2Value();
  for(i=0;i<a_maxDigital;i++){ //デジタルモードで動作中をセットする
    if(digitalPins[i] == cmd_pin) skipmePins[i] = 'D';
  }
  //各値(0,1)に応じて出力する
  if(cmd_value == 0){
    digitalWrite(cmd_pin, LOW); 
  }
  if(cmd_value == 1){
    digitalWrite(cmd_pin, HIGH);
  }
}

//アナログ出力(PWM)
//analogWriteによるデジタルピンのPWM出力
void analogOut()
{
  uint8_t i;
  
  cmd2Value();
  for(i=0;i<a_maxDigital;i++){ //アナログモードで動作中をセットする
    if(digitalPins[i] == cmd_pin) skipmePins[i] = 'A';
  }
  //指定された値を出力する
  analogWrite(cmd_pin, cmd_value);
}

//ピンモード設定(INPUT or OUTPUT)
void pinmSet()
{
  cmd2Value();
  if(cmd_value == 0){
    pinMode(cmd_pin, INPUT);
  }
  if(cmd_value == 1){
    pinMode(cmd_pin, OUTPUT);
  }
}

//コマンド文字列から値へ変換する関数
//コマンドの文字を整数型に変換する
void cmd2Value()
{
  cmd_pin = atoi(strtok(sbuffer+1," "));
  cmd_value = atoi(strtok(NULL," "));
}

//温度と気圧を取得する
void getTempPress()
{
  if(! baro.begin() ){
    my_press = float(0);
    my_temp = float(0);
  } else {
    my_press = baro.getPressure();
    //PaからhPaに変換
    my_press = my_press / (float) 100;
    my_temp = baro.getTemperature();
  }
}

//液晶を使用する場合の初期化
void lcdInit()
{
  lcd.begin(16,2);
  lcd.setBacklight(0x07);
}

//液晶バックライトのオンオフ
//オンオフはX0 X1のかたちになるので
//チェックするのはcmd_pin
void lcdBacklight()
{
  cmd2Value();
  //各値(0,1)に応じてオンオフする
  if(cmd_pin == 0){
    lcd.setBacklight(0x00); 
  }
  if(cmd_pin == 1){
    lcd.setBacklight(0x07);
  }
}

//液晶へのメッセージ表示
void lcdMessage()
{
  String s = sbuffer;
  uint8_t dline,len,len_rest,tmp;
  char tmp_buf[18];

  //受信文字列の2バイト目が表示行
  //各値(1,2)に応じて行を指定する
  if(sbuffer[1] == '1'){
    lcd.setCursor(0,0);
  }
  if(sbuffer[1] == '2'){
    lcd.setCursor(0,1);
  }
  //表示する長さを求める
  len = s.length() - 3;
  if(len >=16) len = 16;
  //後ろにスペースを埋めるので残りの長さを求める
  len_rest = 16 - len;
  //文字列を表示
  lcd.print(sbuffer+3);
  //表示した後ろをスペースで埋める
  if(len_rest > 0){
    for(tmp=0;tmp<len_rest;tmp++){
      tmp_buf[tmp] = 0x20;
      tmp_buf[tmp+1] = 0x00;
    }
    lcd.print(tmp_buf);
  }
}

//サーボピン設定
void setServo()
{
  cmd2Value();

  if(cmd_pin == 1){
    myServo1.attach(cmd_value);
  }
  if(cmd_pin == 2){
    myServo2.attach(cmd_value);
  }
}

//サーボ角度設定
void servoAngle()
{
  cmd2Value();

  if(cmd_pin == 1){
    myServo1.write(cmd_value);
  }
  if(cmd_pin == 2){
    myServo2.write(cmd_value);
  } 
}
